%% tclp__processor.pl clauses to treat files options etc ...
%% to be used with tclp__main.pl and tclp.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                   TCLP                                    %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp.pl,v 1.2 2002/09/04 14:58:49 coquery Exp $

:- module(tclp,[% options
                tclp__set_inference/1,
                tclp__set_functor_labeling/1,
                tclp__add_search_directory/1,
                tclp__set_stdlib_name/1,
                tclp__enable_verbose/1,
                tclp__disable_verbose/1,
                tclp__use_labeling_heuristic/1,
                tclp__labeling_look_ahead/1,
                % processing
                tclp__process_phrases/1,
                tclp__process_pl/1,
                tclp__process_typ/1,
                tclp/1,
                % changing tclp dialect
                tclp__set_dialect/1,
                tclp__reinit/0,
                tclp__reinit/1,
                sicstus/0,
                iso/0,
                gnu/0,
                % help message
                tclp__help/0]).

:- use_module(tclp__utils).
:- use_module(tclp__arrays).
:- use_module(tclp__def).
:- use_module(tclp__arity).
:- use_module(tclp__reader).
:- use_module(tclp__errors).
:- use_module(tclp__type_order).
:- use_module(tclp__clp_types).
:- use_module(tclp__handle_goals).
:- use_module(tclp__declarations).
:- use_module(tclp__files).
:- use_module(tclp__functor_types).
:- use_module(tclp__type_checker).
:- use_module(tclp__connexity).
:- use_module(tclp__type_inference).
:- use_module(sicstus_modules).

:- use_module(library(lists)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Some options
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% enable/disable type inference
:- dynamic do_inference/0, stdlib/1.

tclp__set_inference(DoIt) :-
        (
          DoIt
        ->
          asserta(do_inference)
        ;
          retractall(do_inference)
        ).


%% do a labeling on overloaded symbols
tclp__set_functor_labeling(DoIt) :-
        set_label_functors(DoIt).


%% add a directory to the search path
tclp__add_search_directory(Dir) :-
        add_include(Dir).

%% sets the name of the standard library
%% change the stdlib
change_stdlib(NewLib) :-
        retractall(stdlib(_)),
        asserta(stdlib(NewLib)).

:- change_stdlib('stdlib.typ').

tclp__set_stdlib_name(Name) :-
        change_stdlib(Name).


%% sets verbose levels
tclp__enable_verbose(Level) :-
        enable_verbose(Level,
                       (Level\=types,Level\=choice_points)
                      %% no comment on types nor choice points
                      ),!,
        verbose(always,['Set verbose : ',Level,' on']).
tclp__disable_verbose(Level) :-
        verbose(always,['Set verbose : ',Level,' off']),
        disable_verbose(Level).


%% heuristic for labeling functors
tclp__use_labeling_heuristic(DoIt) :-
        use_heuristic(DoIt).


%% look ahead possibilities
tclp__labeling_look_ahead(LA) :-
        set_look_ahead(LA).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Computing connex parts
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% builds a list of ordered connex parts from a list of phrases
build_connex_parts(PhraseList, OrderedPhrases) :-
        build_ordered_connex_part_list(PhraseList, OrderedPhrases).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Processing to the type checking / type inference
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Type checks one connex part
process_one_connex_part(ConnexPart) :-
        verbose(debug_main,'starting connex part ...'),
        (
          do_inference
        ->
          verbose(debug_main, 'inference'),
          infer_connex_part(ConnexPart)
        ;
          (
            verbose(debug_main, 'type check'),
            type_check_clauses(ConnexPart),!
          ;
            print_error('type check failed')
          )
        ),
        verbose(debug_main,'connex part ok').
process_one_connex_part(_) :-
        print_error('Connex part failed').

%% Type check all connex parts
process_connex_parts([]).
process_connex_parts([ Part | Next ]) :-
        process_one_connex_part(Part),
        verbose(debug_main,['\nProcessing :\n',Part,'\nOK\n']),
        process_connex_parts(Next),
        verbose(debug_main,['\nProcessing :\n',Part]).

process_phrases(Phrases) :- do_inference,!,
        once(chrono('building of connex parts',
                    build_connex_parts(Phrases, OrderedPhrases))),
        once(chrono('computing type order', build_order)),
        once(chrono('type checking',process_connex_parts(OrderedPhrases))).
process_phrases(Phrases) :- 
        once(chrono('computing type order', build_order)),
        once(chrono('type checking',type_check_clauses(Phrases))).

tclp__process_phrases(Phrases) :- process_phrases(Phrases).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Process files
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% read all files and then proceed to the type check of all files at once

%% if there is no files nor phrases do nothing
process_all_files([],[]) :- !.
%% all files where read: PhraseList is the list of all prolog phrases of
%% the files: we proceed to the type check
process_all_files([],PhraseList) :-
        chrono(actions,'checking type declarations',check_declarations),
        process_phrases(PhraseList).
%% we add the phrases of one file
process_all_files([FileName|Rest],CurrentPhraseList) :-
        read_file(FileName, ThisPhraseList),
        append(ThisPhraseList,CurrentPhraseList,PhraseList),
        process_all_files(Rest,PhraseList).

%% Process all prolog files in list
process_pl(FileList) :-
        process_all_files(FileList,[]).

tclp__process_pl(FileList) :- process_pl(FileList).

%% Load all typ files in the list
process_typ([]).
process_typ([ File | Rest ]) :-
        load_file(File),
        process_typ(Rest).

tclp__process_typ(FileList) :- process_typ(FileList).

%% sorts between .typ files and .pl files
filter_typ([],[],[]).
filter_typ([File|Rest],PL,TYP) :-
        (
          atom_concat(_,'.typ',File)
        ->
          PL=PL2 , TYP = [File|TYP2]
        ;
          PL = [File|PL2] , TYP=TYP2
        ),
        filter_typ(Rest,PL2,TYP2).

tclp([]).
tclp([File|Files]) :-
        filter_typ([File|Files],PL,TYP),
        (
          TYP \= [] ->
          once(chrono(['loading : ',TYP,' ...'],process_typ(TYP)));
          true
        ),
        process_pl(PL).
tclp(File) :-
        tclp([File]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% prolog dialects
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic user:tclp__dialect/1,user:init_dialect/1.

get_dialect(Dialect) :-
        user:tclp__dialect(Dialect),!.
get_dialect('ISO').

tclp__set_dialect(Dialect) :-
        asserta(user:tclp__dialect(Dialect)).

user:init_dialect('ISO') :- !,
        tclp__add_search_directory(library('tclp/lib/ISO')),
        tclp__set_stdlib_name('stdlib.typ').
user:init_dialect(sicstus) :- !,
        tclp__add_search_directory(library('tclp/lib/sicstus')),
        tclp__set_stdlib_name('stdlib.typ'),
        enable_sicstus.
user:init_dialect(gnu) :- !,
        tclp__add_search_directory(library('tclp/lib/gprolog')),
        tclp__set_stdlib_name('stdlib.typ').
user:init_dialect(_).

%% reloads definitions for prolog dialects
%% erases all previous definitions

tclp__reinit_modules :-
        tclp__arity:reinit,
        tclp__type_order:reinit,
        types_reinit,
        tclp__handle_goals:reinit,
        tclp__declarations:reinit,
        tclp__files:reinit,
        multi_functor_reinit,
        var_type_reinit.

tclp__reinit :-
        tclp__reinit((get_dialect(Dialect),
                      write('%% dialect is '),write(Dialect),nl,
                      user:init_dialect(Dialect))).

tclp__reinit(Goal) :-
        tclp__reinit_modules,
        Goal,
        stdlib(StdLib),
        once(chrono([ 'loading standard type library (',StdLib,')' ],
                    process_typ([StdLib]))),
        write('%% TCLP ready'),nl.

sicstus :-
        asserta(user:tclp__dialect(sicstus)),
        tclp__reinit.

iso :-
        asserta(user:tclp__dialect('ISO')),
        tclp__reinit.

gnu :-
        asserta(user:tclp__dialect(gnu)),
        tclp__reinit.
%:- tclp__reinit.


%% default options
:-
        tclp__enable_verbose(types),
        tclp__set_inference(true),
        tclp__set_functor_labeling(true).

%:-
%        iso.

tclp__help :-
        write('Available predicates in module tclp:'),nl,
        nl,write('options:'),nl,
        write('  tclp__set_inference(Val)'),nl,
        write('     Val can be either ''true'' or ''false'''),nl,
        write('     en(dis)able predicate type inference'),nl,
        write('  tclp__set_functor_labeling(Val)'),nl,
        write('     Val can be either ''true'' or ''false'''),nl,
        write('     en(dis)able labeling of functor for overloaded symbols.'),nl,
        write('     set to false if you want every symbol overload occurence solved only'),nl,
        write('     by the context'),nl,
        write('  tclp__add_search_directory(Dir)'),nl,
        write('     Dir will be added to the directories where tclp search for files'),nl,
        write('  tclp__set_stdlib_name(Name)'),nl,
        write('     use Name for the standard type library'),nl,
        write('  tclp__enable_verbose(Name)'),nl,
        write('     tclp will print information w.r.t. Name'),nl,
        write('     possible values : types, actions, timing, defaults'),nl,
        write('  tclp__disable_verbose(Name)'),nl,
        write('     tclp will no more print information w.r.t. Name'),nl,
        write('     possible values : types, actions, timing, defaults'),nl,
%                tclp__use_labeling_heuristic/1,
%                tclp__labeling_look_ahead/1,
        nl,write('processing:'),nl,
        write('  tclp__process_phrases(PhraseList)'),nl,
        write('     PhraseList is a list of prolog phrases represented as terms to be'),nl,
        write('     type checked by tclp'),nl,
        write('  tclp__process_pl(PlFiles)'),nl,
        write('     PlFiles is a list of files to be type checked by tclp'),nl,
        write('  tclp__process_typ(TypFiles)'),nl,
        write('     TypFiles is a list of files with type information to load into tclp'),nl,
        write('  tclp(Files)'),nl,
        write('     Files is a list of files to be treated in tclp, .typ files will simply'),nl,
        write('     be loaded, other files will be type checked.'),nl,
        nl,write('changing tclp definitions:'),nl,
%        write('  tclp__set_dialect(Dialect)'),nl,
%        write('     change the name of the used prolog dialect to Dialect. This change
        write('  tclp__reinit'),nl,
        write('     reinits tclp with the definitions of the current standard library'),nl,
%                tclp__reinit/1,
        write('  sicstus'),nl,
        write('     set tclp to type check sicstus prolog programs'),nl,
        write('  iso'),nl,
        write('     set tclp to type check ISO prolog programs'),nl,
        write('  gnu'),nl,
        write('     set tclp to type check GNU prolog programs'),nl,
        nl,write('other:'),nl,
        write('  tclp__help'),nl,
        write('     prints this help message'),nl.
